import configparser
import os
import re
import sqlite3

import pytest

from datetime import datetime
from testrail_api import TestRailAPI

from testrail import TESTRAIL_PREFIX, PYTEST_TESTRAIL_PARAMETRIZE, STATUSES

API = None
RUN_ID = None
CASE_IDS = None


CONFIG_STRUCTURE = [
    {
        'section': 'API',
        'required': True,
        'items': [
            {'key': 'url', 'type': str, 'required': True},
            {'key': 'email', 'type': str, 'required': True},
            {'key': 'password', 'type': str, 'required': True},
            {'key': 'timeout', 'type': float, 'required': False},
        ]
    },
    {
        'section': 'TEST_RUN',
        'required': True,
        'items': [
            {'key': 'project_id', 'type': int, 'required': True},
            {'key': 'assignedto_id', 'type': int, 'required': True},
            {'key': 'name', 'type': str, 'required': False},
        ]
    },
    {
        'section': 'TEST_STATUSES',
        'required': False,
    },
    {
        'section': 'TEST_RESULT',
        'required': False,
    },
]


class ConfigError(Exception):
    pass


@pytest.fixture(scope='function')
def add_result(request):
    """
    Добавить результат выполнения теста (общий метод)
    :param request: pytest Request
    :param kwargs: словарь результатов теста
    """

    def _result(**kwargs):
        if hasattr(request.node, TESTRAIL_PREFIX):
            request.node.testrail.update(kwargs)
        setattr(request.node, TESTRAIL_PREFIX, kwargs)

    return _result


@pytest.fixture(scope='function')
def add_comment(add_result):
    """
    Добавить результат выполнения теста, только в поле custom_comment_custom
    :param add_result: фикстура
    """

    def _comment(message):
        if not message:
            message = 'empty'
        return add_result(custom_comment_custom=message)

    return _comment


def i_am_master(config):
    """
    Метод для определения мастер-воркера при запуске тестов параллельно
    :param config: pytest Config
    :return: True - мастер-воркер, False - слейв-воркер
    """
    return config.workerinput['workerid'] in ('gw0', 'master') if hasattr(config, 'workerinput') else True


class database(object):
    """
    Класс для работы с БД Sqlite
    """

    def __init__(self, config):
        self.connect = sqlite3.connect(os.path.join(config.rootdir.strpath, f'{TESTRAIL_PREFIX}.db'))
        self.connect.row_factory = sqlite3.Row

    def __enter__(self):
        self.cursor = self.connect.cursor()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.cursor.close()
            self.connect.close()
        except Exception:
            pass


def clean_test_ids(test_ids):
    """
    Удаление буквенных префиксов, чтобы получился числовой идентификатор кейса TestRail
    :param list test_ids: список кейсов
    :return list ints: список идентификаторов кейсов
    """

    def _format_case_id(row):
        try:
            if isinstance(row, int):
                return row
            return int(re.search('(?P<test_id>[0-9]+$)', row).groupdict().get('test_id'))
        except Exception:
            return None

    return [_format_case_id(test_id) for test_id in test_ids if _format_case_id(test_id)]


def get_test_cases(item):
    """
    Получить список идентификаторов тест-кейсов, относящихся к тестовой функции item
    :param item: тестовая функция
    :return: список идентификаторов кейсов TestRail
    """

    def _get_test_cases(value):
        cases = []
        if isinstance(value, (tuple, list)):
            for index, row in enumerate(value):
                cases.extend(_get_test_cases(row))
        elif isinstance(value, dict):
            for k, v in value.items():
                if k.lower() in (PYTEST_TESTRAIL_PARAMETRIZE, TESTRAIL_PREFIX):
                    if isinstance(v, (str, int)):
                        cases.append(v)
                    elif isinstance(v, (tuple, list)):
                        cases.extend(v)
                else:
                    cases.extend(_get_test_cases(v))
        else:
            return []
        return cases

    item_cases = []
    marker = item.get_closest_marker(TESTRAIL_PREFIX)
    if marker:
        item_cases.extend(marker.args)
    markers = [marker for marker in item.iter_markers('parametrize')]
    if markers:
        current_params = item.callspec.params
        for current_key, current_value in current_params.items():
            if isinstance(current_value, (dict, list, tuple)):
                item_cases.extend(_get_test_cases(current_value))
            elif isinstance(current_value, str):
                for marker in markers:
                    if current_key == marker.args[0] and isinstance(marker.args[1], dict):
                        item_cases.extend(_get_test_cases(marker.args[1][current_value]))
        item_cases.extend(_get_test_cases(item.callspec.params))
    return clean_test_ids(item_cases)


def check_testrail_config(cfg_path):
    """
    Функция для проверки конфига TestRail
    :param cfg_path: путь до файла конфига TestRail
    :return: словарь ключ-значение конфига TestRail
    """
    if os.path.isfile(cfg_path):
        cfg = configparser.ConfigParser()
        cfg.read(cfg_path)
        sections = cfg.sections()
    else:
        raise ConfigError(f'Не найден конфиг для TestRail в: {cfg_path}')

    cfg_params = {s.lower(): dict(cfg[s]) for s in sections}

    required_sections = [row['section'] for row in CONFIG_STRUCTURE if row['required']]
    if not set(required_sections).issubset(set(required_sections)):
        raise ConfigError(f'Невалидный файл TestRail конфига. Не найдено секций: {required_sections}')

    for row in CONFIG_STRUCTURE:
        for item in row.get('items', []):
            key_name = item['key']
            value_type = item['type']
            key_required = item['required']
            try:
                value = value_type(cfg_params[row['section'].lower()][key_name])
                cfg_params[row['section'].lower()][key_name] = value
            except KeyError:
                if key_required:
                    raise ConfigError(f'Не найден обязательный параметр TestRail конфига: {key_name}')
            except (TypeError, ValueError):
                raise ConfigError(f'Невалидное значение у параметра конфига: {key_name} ожидается {value_type}')

    test_run_name = cfg_params.get('test_run', {}).get('name')
    if test_run_name:
        cfg_params['test_run']['name'] += f' {datetime.today().strftime("%Y-%m-%d %H:%M:%S")}'
    return cfg_params


def get_test_result_for_report(item):
    """
    Получить результаты теста
    :param item: тест-функция
    """
    # получить отчет о прохождении теста или setup части, если до выполнения теста не дошло
    repoprt = getattr(item, 'rep_call', None) or getattr(item, 'rep_setup', None)

    # шаблон для отправки результата теста, по умолчанию берутся значения из конфига testrail.cfg - секция TEST_RESULT
    test_result_template = dict(**getattr(item.config, TESTRAIL_PREFIX, {}).get('test_result', {}))

    # получить статус выполнения теста из отчета
    status_id = STATUSES.get(repoprt.outcome, 4)

    # вычислить время выполнения теста, TestRail не принимаетвремя меньше 1 секунды
    elapsed_time = f'{1 if repoprt.duration < 1 else int(round(repoprt.duration))}s'

    # заполнить поле Comment, если тест не passed, то будут добавлены логи выполнения теста
    comment = f'[{repoprt.outcome.upper()}]\n'
    if status_id != STATUSES['passed']:
        comment += repoprt.capstdout if repoprt.capstdout else repoprt.caplog if repoprt.longrepr else repoprt.outcome

    # остальные поля, которые можно передать через фикстуру add_result или add_comment
    other = getattr(item, TESTRAIL_PREFIX, {})

    test_result_template.update(
        status_id=status_id,
        elapsed=elapsed_time,
        comment=comment,
        **other,
    )
    return test_result_template


def pytest_addoption(parser):
    """
    Включает интеграцию с TestRail опцией --testrail.
    Опционально:
        --testrail-config - путь до конфига TestRail
        --testrail-run-id - идентификатор тест-рана
        --testrail-run-title - заголовок тест-рана
    """
    group = parser.getgroup(TESTRAIL_PREFIX)

    group.addoption(f'--{TESTRAIL_PREFIX}', action='store_true', help='')
    group.addoption(f'--{TESTRAIL_PREFIX}-config', action='store', help='')
    group.addoption(f'--{TESTRAIL_PREFIX}-run-id', action='store', help='')
    group.addoption(f'--{TESTRAIL_PREFIX}-run-title', action='store', help='')


def pytest_report_header(config):
    """
    Заголовок при выполнении тестов с интеграции TestRail и проверка наличия конфига,
    достаточности параметров в конфиге. По умолчанию конфиг должен лежать в корне проекта config/testrail.cfg
    :param config: pytest Config
    :return: str
    """
    if config.getoption(f'--{TESTRAIL_PREFIX}'):
        return 'Thanks for using pytest-testrail-api plugin.'


def pytest_configure(config):
    config.addinivalue_line("markers", "testrail: Маркер для обозначения кейсов в TestRail")


@pytest.hookimpl(trylast=True)
def pytest_collection_modifyitems(session, config, items):
    if config.getoption(f'--{TESTRAIL_PREFIX}'):
        global RUN_ID
        global CASE_IDS
        global API

        # получить путь до конфиг-файла с настройками TestRail
        testrail_cfg = config.getoption(
            f'--{TESTRAIL_PREFIX}-config',
            default=os.path.join(config.rootdir.strpath, 'config', f'{TESTRAIL_PREFIX}.cfg')
        )

        # проверка конфига и добавление параметров в pytest Config
        params = check_testrail_config(testrail_cfg)
        setattr(config, TESTRAIL_PREFIX, params)

        # обновление статусов результатов теста TestRail
        STATUSES.update(params.get('test_statuses', {}))

        API = TestRailAPI(**params['api'])

        # выполнение только мастер-процессом
        if i_am_master(config):
            with database(config) as db:
                # удаление данных из таблиц, если они существуют
                db.cursor.execute('DROP TABLE IF EXISTS run;')
                db.connect.commit()
                db.cursor.execute('DROP TABLE IF EXISTS cases;')
                db.connect.commit()

                # создание таблиц, если их нет
                db.cursor.execute('CREATE TABLE IF NOT EXISTS run(id INT);')
                db.connect.commit()
                db.cursor.execute('CREATE TABLE IF NOT EXISTS cases(id INT);')
                db.connect.commit()

            RUN_ID = config.getoption(f'--{TESTRAIL_PREFIX}-run-id', default=None)
            run_title = config.getoption(
                f'--{TESTRAIL_PREFIX}-run-title',
                default=f'Протокол автоматизированного тестирования {datetime.today().strftime("%Y-%m-%d %H:%M:%S")}')
            if RUN_ID is None:
                CASE_IDS = []
                for item in items:
                    CASE_IDS.extend(get_test_cases(item))
                CASE_IDS = list(set(CASE_IDS))
                # создать новый тест-ран в TestRail, если не передан идентификатор тест-рана
                new_run = dict(name=run_title, include_all=False, case_ids=CASE_IDS)
                new_run.update(getattr(config, TESTRAIL_PREFIX, {}).get('test_run', {}))

                r = API.runs.add_run(**new_run)

                RUN_ID = int(r['id'])

            # получить список тест-кейсов и список кейсов по идентификатору тест-рана, если он передан
            tests = API.tests.get_tests(run_id=RUN_ID)
            CASE_IDS = [int(test['case_id']) for test in tests]

            # сохранение в БД нужно для простого шаринга данных между параллельными процессами pytest, причем
            # запись данных осуществляется только мастер-процессом
            with database(config) as db:
                # заполнение таблиц данными, необходимыми всем параллельным процессам
                db.cursor.execute(f'INSERT INTO run (id) VALUES ({RUN_ID})')
                db.connect.commit()
                if CASE_IDS:
                    cases = [(case,) for case in CASE_IDS]
                    db.cursor.executemany('INSERT OR REPLACE INTO cases VALUES (?)', cases)
                    db.connect.commit()
        else:
            while True:
                with database(config) as db:
                    if db.cursor.execute('SELECT * FROM run').fetchone():
                        break


@pytest.hookimpl(hookwrapper=True, tryfirst=True)
def pytest_runtest_makereport(item, call):
    """
    Получение и сохранение результатов прохождения setup/теста/teardown частей
    :param item: тест-функция
    """
    outcome = yield
    rep = outcome.get_result()
    setattr(item, f'rep_{rep.when}', rep)


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_setup(item):
    """
    Hook - перед setup частью теста. Поиск идентификаторов кейсов TestRail в pytest.mark.testrail/parametrize.
    Проверка на пересечения со списком кейсов из тест-рана. Если кейсы из тест-рана есть, а кейсов из теста нет, или
    нет пересечений - такой тест помечатеся pytest.mark.skip для пропуска теста. В дургом случае тест запускается.
    :param item: тест-функция
    """
    if item.config.getoption(f'--{TESTRAIL_PREFIX}'):
        item_cases = get_test_cases(item)
        global CASE_IDS
        if CASE_IDS is None:
            with database(item.config) as db:
                cases = db.cursor.execute('SELECT * FROM cases').fetchall()
                CASE_IDS = set([case['id'] for case in cases])
        if item_cases and CASE_IDS:
            if not set(item_cases).intersection():
                item.add_marker(pytest.mark.skip('Test not present in Test Run. Skip!'))
        elif CASE_IDS:
            item.add_marker(pytest.mark.skip('Test not marked as testrail case. Skip!'))


@pytest.hookimpl(trylast=True)
def pytest_runtest_teardown(item):
    """
    Hook - после teardown части теста. Отправка результатов теста в тест-ран.
    :param item: тест-функция
    """
    if item.config.getoption(f'--{TESTRAIL_PREFIX}'):
        item_cases = get_test_cases(item)
        global RUN_ID
        if RUN_ID is None:
            with database(item.config) as db:
                run = db.cursor.execute('SELECT * FROM run').fetchone()
                RUN_ID = run['id']
        for case in item_cases:
            API.results.add_result_for_case(
                run_id=RUN_ID,
                case_id=case,
                **get_test_result_for_report(item)
            )
